@@ -1,11 +1,25 @@
 package SDCommon;
 
+#  Copyright 2003, 2004, 2005, 2006 Eduard Bloch <blade@debian.org>
+#
+#  This package is free software; you can redistribute it and/or modify
+#  it under the terms of the GNU General Public License as published by
+#  the Free Software Foundation; either version 3 of the License, or
+#  (at your option) any later version.
+#
+#  This program is distributed in the hope that it will be useful,
+#  but WITHOUT ANY WARRANTY; without even the implied warranty of
+#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#  GNU General Public License for more details.
+#
+#  You should have received a copy of the GNU General Public License
+
 use base qw/ Exporter /;
 our @EXPORT = qw(writeCfg withecho withechoNoPrompt needs_upsTagUrl
 needs_upsCurrentUrl needs_tagsUrl url insvn long_path oldSvnDirsCheck
-load_dirs set_statusref);
+load_dirs set_statusref svnMkdirP start_ssh);
 #### common ####
-our @EXPORT_OK = qw ($tagVersion $upVersion $package $version $opt_verbose $opt_noninteractive %c $nosave $opt_quiet $epoch $version);
+our @EXPORT_OK = qw ($tagVersion $upVersion $package $version $opt_verbose $opt_noninteractive %c $opt_nosave $opt_quiet $epoch $version);
 use vars qw($package $epoch $version $opt_verbose);
 use Cwd;
 use SVN::Client;
@@ -19,15 +33,22 @@ use strict;
 #my $upVersion;
 #my $package;
 #my $epoch;
-#my $nosave;
+my $opt_nosave;
 my @cfgRest;
 my %cacheUrl;
 my %inSvn;
 my @junk;
 my $force_debian;
 
-# to be replaced during the package creation
-$SDCommon::version=0.6.21;
+# static fallback value
+my $OURVERSION = '0.6.25';
+
+sub common_version {
+	my $query = `dpkg-query -W -f='\${Version}' svn-buildpackage`;
+	(defined $query) ? return $query : return $OURVERSION;
+}
+
+$SDCommon::version=&common_version();
 
 # workaround for the old piece of Perl shit in Woody
 sub long_path {
@@ -71,8 +92,10 @@ sub withecho {
    retry:
    print STDERR "$cmd\n" if(!$SDCommon::opt_quiet);
    if(system(@_)) {
+      my $cdir = Cwd::getcwd ;
+      $cdir = "<unknown>" unless ($cdir);
       print STDERR "Command $cmd failed in ".
-         ($Cwd::getcwd?$Cwd::getcwd:"<unknown>").
+         $cdir.
          ", how to continue now? [Qri?]: ";
       prompt:
       my $ans = <STDIN>;
@@ -110,8 +133,33 @@ sub withechoNoPrompt {
       $cmd.=" <$rest more argument".($rest>1?"s>":">") if $rest;
    }
    print STDERR "$cmd\n" if(!$SDCommon::opt_quiet);
-   return 1 if ($SDCommon::opt_ignoreerrors);
-   return (!system(@_));
+   return ( (!system(@_)) or ($SDCommon::opt_ignoreerrors) );
+}
+
+sub svnMkdirP {
+   my $url=shift;
+   my $scriptname = shift;
+   my $tempurl=$url;
+   my $missingurl="";
+   my $part;
+   my $noexist=0;
+   my $svnout;
+   my $svntempdir;
+
+   chomp ( $svnout = `env LC_ALL=C svn info $tempurl 2>/dev/null`) ;
+   while ( ( $svnout =~ /^$/ ) and ( "$tempurl" =~ /^(svn(\+ssh)?|file|https?):\/\/.*/ ) ) {
+      chomp ( $part= `basename $tempurl` ) ;
+      $missingurl = $part . '/' . $missingurl ;
+      chomp ( $tempurl = `dirname $tempurl` ) ;
+      chomp ( $svnout = `env LC_ALL=C svn info $tempurl 2>/dev/null`) ;
+   } ;
+
+   # now we have to import $missingurl at $tempurl level
+   return 0 if ($missingurl =~ /^$/ ) ;
+   chomp ( $svntempdir=`mktemp -d` ) ;
+   system ( "mkdir -p $svntempdir/$missingurl" ) ;
+   withecho("svn", "import", "$svntempdir", "$tempurl", "-m", "$scriptname Creating $missingurl directory." );
+   system "rm -fr $svntempdir 2>/dev/null" ;
 }
 
 sub oldSvnDirsCheck {
@@ -130,7 +178,7 @@ sub oldSvnDirsCheck {
       exit 1;
    }
 }
-         
+
 
 # gets the url of specified local (checkout) directory
 sub url {
@@ -141,7 +189,7 @@ sub url {
    my $PATH;
    my $testpath=long_path($_[0]);
    return $cacheUrl{$testpath} if($cacheUrl{$testpath});
-   open(INFO, "svn info $testpath 2>/dev/null |");
+   open(INFO, "env LC_ALL=C svn info $testpath 2>/dev/null |");
    while(<INFO>) { 
       $URL=$1 if(/^Url\s*:\W*(.+)\n/i); 
       $PATH=$1 if(/^Path\s*:\s*(.+)\n/i);
@@ -171,7 +219,7 @@ sub insvn {
    return undef if !defined $testurl;
    return  $inSvn{$testurl} if $inSvn{$testurl};
    print "Repository lookup, probing $url...\n";
-   open(INFO, "svn ls $url 2>/dev/null |") or
+   open(INFO, "env LC_ALL=C svn ls $url 2>/dev/null |") or
       die ("Can't open svn ls $url: $!");
    @junk=<INFO>;
    if (close INFO) {
@@ -183,8 +231,8 @@ sub insvn {
 
 # helper to automate lookup for variables
 # does not change defined var, but looks for useful path if not defined
-sub such {
-   
+sub search {
+
    my @testloc;
    my @testsvn;
    our($pre, $suf, $var, $basedir) = @_;
@@ -205,12 +253,21 @@ sub such {
    # basedir: starting directory. If ommited, `pwd` is used
 
    return if (defined($SDCommon::c{$var}));
-   
+
+   # FIXME: this should use a smarter way of detecting an url,
+   # see 433532 for more details
    $basedir=Cwd::getcwd if(!$basedir);
    if($basedir=~/:\/\//) { # huch, URL was specified as $basedir?
       print "W: $var not specified anywhere, looking in the local repository...\n";
       @testsvn=("$basedir/../$pre/$suf",
-      "$basedir/../../$pre/$package/$suf");
+      "$basedir/../../$pre/$package/$suf",
+      "$basedir/../../$pre/$suf/$package");
+      if($basedir =~ /(trunk|branches|tags)/ && $pre =~ /^(branches|tags|trunk)$/) {
+        my $simple = "$basedir";
+        $simple =~ s/(trunk|branches|tags)/$pre/;
+        $simple .= "/$suf" if(length($suf));
+        unshift @testsvn, $simple;
+      }
    }
    else {
       if($SDCommon::opt_verbose) {
@@ -226,7 +283,7 @@ sub such {
    for (@testsvn) {
      print "Looking in SVN for: @testsvn\n";
       if ($_ && insvn($_)) { 
-         print "I: adding the URLs to the $cfgFile to skip the check later.\n";
+         print "I: adding the URLs to the $cfgFile to skip the check later.\n" if(!$SDCommon::opt_nosave);
          $SDCommon::c{$var}=insvn($_);
          return;
       }
@@ -238,24 +295,31 @@ sub such {
 
 # if still undefined, try blind search in the repository
 sub needs_upsTagUrl {
-   such("branches","upstream", "upsTagUrl", url("."));
+   search("branches","upstream", "upsTagUrl", url("."));
    exwerror "upsTagUrl setting is required, but could not be found in $cfgFile or repository tree." if (!defined $SDCommon::c{"upsTagUrl"});
 }
 sub needs_upsCurrentUrl {
-   such("branches","upstream/current", "upsCurrentUrl", url("."));
+   search("branches","upstream/current", "upsCurrentUrl", url("."));
+   if( !defined $SDCommon::c{"upsCurrentUrl"} ) { # maybe layout 2?
+        search("branches/upstream","current", "upsCurrentUrl", url("."));
+   }
    exwerror "upsCurrentUrl setting is required, but could not be found in $cfgFile or repository tree." if (!defined $SDCommon::c{"upsCurrentUrl"});
 }
 sub needs_tagsUrl {
     my $hereurl=url(".");
-    such("tags", "", "tagsUrl", $hereurl);
+    search("tags", "", "tagsUrl", $hereurl);
     if( !defined $SDCommon::c{"tagsUrl"} && $hereurl=~/(.*)\/branches\// ) { # oh, crap...
         print "Branch directory detected in URL, testing more possible locations\n";
-        such("tags", "", "tagsUrl", "$1/$package");
+        search("tags", "", "tagsUrl", "$1/$package");
     }
     exwerror "tagsUrl setting is required, but could not be found in $cfgFile or repository tree." if (!defined $SDCommon::c{"tagsUrl"});
 }
 sub writeCfg {
+   my $cfgFile = shift;
+   print "I: Configuration saving was overrided ($cfgFile should be intact).\n" if ($SDCommon::opt_nosave==1 and (defined $SDCommon::opt_verbose) );
+   return 1 if $SDCommon::opt_nosave==1;
    if($_[0]){$cfgFile = $_[0];}
+   print "I: Trying to save the configuration to $cfgFile.\n" if defined $SDCommon::opt_verbose;
    open(CFG, ">$cfgFile") || die "Could not open $cfgFile for writing.\n";
    foreach (sort(keys %SDCommon::c)) {
       print CFG "$_=".$SDCommon::c{$_}."\n" if defined $SDCommon::c{$_};
@@ -265,11 +329,10 @@ sub writeCfg {
 }
 
 sub sd_exit {
-  my $ret;
-   if(!$SDCommon::nosave) {
+   my $ret = shift||0;
+   if($SDCommon::opt_nosave==0) {
       print "Writing config: $cfgFile\n\n\n" if $SDCommon::opt_verbose;
-      $ret=$_[0];
-      &writeCfg();
+      &writeCfg(".svn/deb-layout");
    }
    # print STDERR "Return-Code: $ret";
    exit $ret;
@@ -278,7 +341,7 @@ sub sd_exit {
 sub init {
 
 
-   open(INFOC, "svn info debian/changelog |");  @junk=<INFOC>;
+   open(INFOC, "env LC_ALL=C svn info debian/changelog |");  @junk=<INFOC>;
    if(!close(INFOC)) {
       exwerror "Not started from the Trunk directory or not a valid SVN repository. Aborting.\n";
    }
@@ -304,7 +367,10 @@ sub init {
       $package=$1;
       $SDCommon::upVersion=$4;
       $epoch=$3;
-      $SDCommon::tagVersion=$SDCommon::upVersion;
+      $SDCommon::tagVersion="$SDCommon::upVersion";
+      if(defined($epoch) && length($epoch)) {
+        $SDCommon::tagVersion = "$epoch:".$SDCommon::tagVersion;
+      }
       if(!$force_debian) {
          $SDCommon::upVersion=~s/^.*://;
          $SDCommon::upVersion=~s/(.*)-([^-]+)/$1/;
@@ -322,51 +388,111 @@ sub init {
 
    " if $SDCommon::opt_verbose;
 
+   $SDCommon::opt_nosave=1;
+
+}
+
+# File::Spec's catfile will break file:// and svn+ssh:// because it
+# removes the double /
+sub mycatfile {
+  my $str = "";
+  my $i = 0;
+  foreach(@_) {
+    if(!($i == 0 or substr($_, 0, 1) eq "/" or substr($str, -1) eq "/")) {
+      $str .= "/";
+    }
+    $str .= $_;
+    $i++;
+  }
+  return $str;
+}
+
+sub readLayoutProps {
+   my $nonempty = 0 ;
+   my $repo_root;
+   foreach ( `svn proplist debian | grep 'svn-bp:'` ) {
+   # import every svn-bp:* property as a cfg
+   if(/\s*svn-bp:(\S+)\s*/) {
+      my $prop = $1;
+      my $val=`svn propget svn-bp:$prop debian | head -n1 | tr -d '\n'`;
+      $val=~ s/\ ~/\ $ENV{"HOME"}/;
+      if(($val !~ m%://%) and ($prop =~ /Url/)) {
+        if(!defined($repo_root)) {
+          $repo_root=`env LC_ALL=C svn info | grep "^Repository Root: " | cut -d ' ' -f 3-`;
+          chomp $repo_root;
+        }
+        $val = mycatfile($repo_root, $val);
+      }
+      # make sure we don't override anything from the command line
+      # or some other prioritary levels
+      $SDCommon::c{$prop}=$val if (!defined $SDCommon::c{$prop});
+      print "Importing layout information via Subversion properties... \n" if ( $nonempty == 0 );
+      print "\t$prop: $val\n";
+      $nonempty = 1 ;
+      }
+   }
+}
+
+sub printLayout {
+   our ($msg) = @_;
+   my $nonempty = 0 ;
+   $msg = "Layout info" if ( !defined $msg ) ;
+   print "$msg:\n" ;
+   foreach (sort(keys %SDCommon::c)) {
+      print "\t$_=".$SDCommon::c{$_}."\n" if (defined $SDCommon::c{$_}) ; #if($_=~/^\// && ! -e $_);
+      $nonempty = 1;
+      };
+   print "\t(no information)\n" if ( $nonempty == 0 ) ;
+}
+
+sub readLayoutFromFile {
+   our ($layoutFile, $formalFileName) = @_;
+
+   if(! -e $layoutFile) {
+      print "No layout information imported through $formalFileName.\n$layoutFile does not exist.\n" if ( defined $opt_verbose ) ;
+   } else {
+      print "Importing layout information from $layoutFile.\n" if ( defined $opt_verbose ) ;
+      if(-r $layoutFile) {
+         open(CFG, "<$layoutFile") or die ("Can't open $layoutFile: $!");
+         while(<CFG>) {
+            if(/(\S+)\s*=\s*(.+)(\n|\r)*/) {
+               my $val=$2;
+               $val=~ s/\ ~/\ $ENV{"HOME"}/;
+               $SDCommon::c{$1}=$val;
+               print "\t$1: $val\n";
+            }
+            else {
+               push(@cfgRest,$_);
+            }
+         }
+         close(CFG);
+      }
+   };
 }
 
 sub configure {
+  my @opt_override = @_;
 
    &init if(!$SDCommon::tagVersion);
    # keep the list of known vars here
    # @cfgVars=("upsCurrentDir", "upsTagDir", "tagsDir", "origDir", "origUrl",
    # "upsCurrentUrl","upsTagUrl", "tagsUrl", "trunkUrl", "trunkDir");
 
-   my $fromFile=$cfgFile;
-   if(!-e $cfgFile) {
-      if(-e $defCfgFile) {
-         print "$cfgFile not found, importing defaults from $defCfgFile\n";
-         $fromFile=$defCfgFile;
-      }
-      else {
-         print "$cfgFile not found, importing settings via Subversion properties... \n";
-         foreach ( `svn proplist debian | grep 'svn-bp:'` ) {
-            # import every svn-bp:* property as a cfg
-            if(/\s*svn-bp:(\S+)\s*/) {
-               my $val=`svn propget svn-bp:$1 debian | head -n1 | tr -d '\n'`;
-               $val=~ s/\ ~/\ $ENV{"HOME"}/;
-               $SDCommon::c{$1}=$val;
-               print "\t$1: $val\n";
-            }
-         }
-         print "Autodetecting remaining properties... \n";
-      }
-   }
-      
-   if(-r $fromFile) {
-      open(CFG, "<$fromFile") or die ("Can't open $fromFile: $!");
-      while(<CFG>) {
-         if(/(\S+)\s*=\s*(.+)(\n|\r)*/) {
-            my $val=$2;
-            $val=~ s/\ ~/\ $ENV{"HOME"}/;
-            $SDCommon::c{$1}=$val;
-            print "\t$1: $val\n";
-         }
-         else {
-            push(@cfgRest,$_);
-         }
-      }
-      close(CFG);
-   }
+   # import config options in the right order so that the
+   # priority of configuration options is (later is higher):
+   #  props, .svn/deb-layout, command line
+
+   &readLayoutProps;
+   &printLayout ("Versioned layout properties") if (defined $SDCommon::opt_verbose);
+
+   &readLayoutFromFile ( $defCfgFile, "the versioned file layout" ) ;
+   # status of the versioned layout info
+   &printLayout ("Versioned layout information") if (defined $SDCommon::opt_verbose);
+
+   &readLayoutFromFile ( $cfgFile, "the local override" ) ;
+
+   # give a nice status of the currently set options
+   &printLayout ("Layout information with local overrides") if (defined $SDCommon::opt_verbose);
 
    # always redetect them; keep them in the same config base just for the
    # record
@@ -382,24 +508,38 @@ sub configure {
       exwerror "\nThe directory $_ does not exist!
 
 Create this directory or fix the setting in .svn/deb-layout or remove that
-line and let svn-descripts redetect the value. Also check the associated URL.
+line and let svn-buildpackage redetect the value. Also check the associated
+URL.
 
 " if($_=~/^\// && ! -e $_);
    }
    # lookup for local locations and get URLs from them if needed
-   such("build-area", "", "buildArea");
-   such("tags", "", "tagsDir");
+   search("build-area", "", "buildArea");
+   search("tags", "", "tagsDir");
    $SDCommon::c{"tagsUrl"}=url($SDCommon::c{"tagsDir"}) if($SDCommon::c{"tagsDir"} && !defined $SDCommon::c{"tagsUrl"});
    if($SDCommon::upVersion ne $SDCommon::tagVersion) {
-      such("branches","upstream", "upsTagDir");
+      search("branches","upstream", "upsTagDir");
       $SDCommon::c{"upsTagUrl"}=url($SDCommon::c{"upsTagDir"}) if($SDCommon::c{"upsTagDir"} && !defined $SDCommon::c{"upsTagUrl"});
-      such("branches","upstream/current", "upsCurrentDir");
+      search("branches","upstream/current", "upsCurrentDir");
       $SDCommon::c{"upsCurrentUrl"}=url($SDCommon::c{"upsCurrentDir"}) if($SDCommon::c{"upsCurrentDir"} && !defined $SDCommon::c{"upsCurrentUrl"});
-      such("tarballs", "", "origDir");
+      search("tarballs", "", "origDir");
    }
 
-   &writeCfg;
-   #
+  @opt_override = split(/,|\ |\r|\n/,join(',',@opt_override));
+  for(@opt_override) {
+    if(/(.*)=(.*)/) {
+      print "Overriding variable: $1 with $2\n" if (defined $SDCommon::opt_verbose);
+      $SDCommon::c{$1}=$2;
+    }
+    else {
+      print "Warning, unable to parse the override string: $_\n";
+    }
+  }
+
+   &printLayout ("Complete layout information");
+
+   &writeCfg(".svn/deb-layout");
+
    #foreach("branches", "tarballs", "trunk", "tags") {
    #   if(length($pkgDir)) {
    #      exwerror "E: Weird directory structure. Where am I? Missing $startdir/$_/$package/\n"
@@ -416,8 +556,10 @@ line and let svn-descripts redetect the value. Also check the associated URL.
 
 sub check_uncommited {
    my  @conflicts;
-   open(SVN, "svn status |") or die ("Can't open `svn status`: $!");
+   open(SVN, "env LC_ALL=C svn status --ignore-externals |") or die ("Can't open `svn status`: $!");
    for(<SVN>) {
+       next if /^\s*$/;
+       next if /^\s*X+\s+(.*)/;
        if(/^\s*M+\s+(.*)/) {
            # FIXME: rewrite to run svn propget in one command with a list if
            # somebody complains about performance issues
@@ -507,12 +649,85 @@ sub load_dirs {
     for(keys %tmp) {
         if($tmp{$_} == 2) {
             $ctx->add($_, 1);
+        } elsif($tmp{$_} == 11) {
+           if ($SDCommon::opt_ignored_files_action) {
+              exwerror "Action for ignored files must be either 'import' or 'skip'!\nAborting!"
+                unless ($SDCommon::opt_ignored_files_action eq "import" or $SDCommon::opt_ignored_files_action eq "skip");
+              $ctx->add($_, 1) if $SDCommon::opt_ignored_files_action eq "import";
+           } else {
+              unless ($SDCommon::opt_noninteractive) {
+                 print STDERR "One or more files are ignore due to (global) subversion ignore patterns,\n\thow to procede (quit, import, skip)? [Qis?]: ";
+                 prompt:
+                 my $ans = <STDIN>;
+                 exwerror "Aborting!\n" if ($ans =~ /^$/ or $ans =~ /^q/i);
+                 if ($ans =~ /^i/i) {
+                    $ctx->add($_, 1);
+                    $SDCommon::opt_ignored_files_action = "import";
+                 } elsif ($ans =~ /^s/i) {
+                    $SDCommon::opt_ignored_files_action = "skip";
+                 } else {
+                    print STDERR "Your choices are:\n\tQuit (q), import (i) which means to add ignored files or skip (s)\n\twhich means to NOT include those files.\n";
+                    goto prompt;
+                 }
+              }
+           }
         }
-   }
+    }
     chdir $curdir;
 }
 
+our @end_ssh_command;
+our $socket;
+
+use File::Temp qw/tempdir/;
+use File::Spec;
+
+sub start_ssh {
+  my $url = shift;
+  if(!($url)) {
+    $url = @{[grep /^Repository Root: /, `env LC_ALL=C svn info`]}[0];
+    $url =~ s/Repository Root: //;
+    chomp $url;
+  }
+  if(!($url =~ /^svn\+ssh:\/\//)) {
+    return;
+  }
+  $url =~ s/^svn\+ssh:\/\/([^\/]+)\/.*$/$1/;
+  my($user, $host, $port) = ($url =~ /(?:([^@]*)@)?([^:]*)(?:\:(.*))?/);
+  return if(!($host));
+  my @ssh_command = qw/ssh/;
+  push @ssh_command, "-l", $user if($user);
+  push @ssh_command, "-p", $port if($port);
+  push @ssh_command, $host;
+  $socket = File::Spec->catfile(tempdir(CLEANUP => 1), "socket");
+  push @ssh_command, "-S", $socket;
+  my @start_ssh_command = @ssh_command;
+  push @start_ssh_command, "-o", "ControlMaster=yes", "-f", "-N";
+  print "Starting ssh connection..\n";
+  system(@start_ssh_command);
+  my $ret = $? >> 8;
+  @end_ssh_command = @ssh_command;
+  push @end_ssh_command, "-O", "exit";
+  if($ret != 0) {
+    exwerror("Failed to make the ssh connection, ssh's return code was: " . $ret);
+  }
+  my @svn_ssh = qw/ssh/;
+  push @svn_ssh, "-S", $socket;
+  $ENV{SVN_SSH} = join " ", @svn_ssh;
+}
+
+sub stop_ssh {
+  return if(!@end_ssh_command);
+  return if(! -e $socket);
+  system(@end_ssh_command);
+}
+
+sub END {
+  stop_ssh();
+}
+
 1;
 ##
 ##### /common ###
+# vi: se sw=3 ts=3 et: #
 
